import { Action } from 'popmotion';
import { ComponentType } from 'react';
import { CSSProperties } from 'react';
import { DetailedHTMLFactory } from 'react';
import { FrameData } from 'framesync';
import * as React from 'react';
import { ReactHTML } from 'react';
import { ReactNode } from 'react';
import { RefObject } from 'react';
import { SVGAttributes } from 'react';

declare type ActionFactory = (actionConfig: PopmotionTransitionProps) => Action;

declare class AnimationControls<P = {}> {
    private props;
    private values;
    private ref;
    private variants;
    private baseTarget;
    private overrides;
    private activeOverrides;
    private resolvedOverrides;
    private defaultTransition?;
    private children?;
    private isAnimating;
    constructor(values: MotionValuesMap, ref: RefObject<Element>);
    setProps(props: P): void;
    setVariants(variants?: Variants): void;
    setDefaultTransition(transition?: Transition): void;
    setValues(target: Target, isActive?: Set<string>): void;
    checkForNewValues(target: Target): void;
    resolveVariant(variant?: Variant): {
        target?: Target;
        transition?: Transition;
        transitionEnd?: Target;
    };
    getHighestPriority(): number;
    setOverride(definition: AnimationDefinition, overrideIndex: number): void;
    startOverride(overrideIndex: number): Promise<any> | undefined;
    clearOverride(overrideIndex: number): void;
    apply(definition: VariantLabels | TargetAndTransition): void;
    private applyVariantLabels;
    start(definition: AnimationDefinition, opts?: AnimationOptions): Promise<any>;
    isHighestPriority(priority: number): boolean;
    private animate;
    private animateVariantLabels;
    private animateVariant;
    private animateChildren;
    private checkOverrideIsAnimating;
    private resetIsAnimating;
    stop(): void;
    addChild(controls: AnimationControls): void;
    removeChild(controls: AnimationControls): void;
    resetChildren(): void;
}

declare type AnimationDefinition = VariantLabels | TargetAndTransition | TargetResolver;

declare class AnimationGroupControls {
    private hasMounted;
    private defaultTransition;
    private pendingAnimations;
    private componentControls;
    private variants;
    setVariants(variants: Variants): void;
    setDefaultTransition(transition: Transition): void;
    subscribe(controls: AnimationControls): () => boolean;
    start(definition: Variant | string): Promise<any>;
    stop(): void;
    mount(): void;
    unmount(): void;
}

declare type AnimationOptions = {
    delay?: number;
    priority?: number;
};

declare type Config<T> = {
    transformer?: Transformer<T>;
    parent?: MotionValue<T>;
};

declare type Constraints = {
    left?: number;
    right?: number;
    top?: number;
    bottom?: number;
};

declare interface ControlsProp {
    controls?: AnimationControls;
}

declare type CSSPropertiesWithoutTransition = Omit_2<CSSProperties, "transition">;

declare type CubicBezier = [number, number, number, number];

declare type CustomMotionComponent = {
    custom: typeof createMotionComponent;
};

declare type Decay = {
    type: "decay";
    modifyTarget?: (v: number) => number;
    power?: number;
    timeConstant?: number;
    restDelta?: number;
    from?: number | string;
    velocity?: number;
    delay?: number;
};

declare type DragDirection = "x" | "y";

declare interface DraggableProps {
    /**
     * Enable dragging for this element
     *
     * Set "x" or "y" to only drag in a specific direction
     * Set "lockDirection" to lock dragging into the initial direction
     *  @default false
     */
    dragEnabled?: boolean | DragDirection | "lockDirection";
    /**
     * Allows drag gesture propagation to child components.
     * @default false
     */
    dragPropagation?: boolean;
    /**
     * An object of optional `top`, `left`, `right`, `bottom` pixel values, beyond which dragging is constrained
     * @default undefined
     */
    dragConstraints?: Constraints;
    /**
     * The degree of movement allowed outside constraints. Set to `false` for no movement.
     * @default 0.5
     */
    dragElastic?: Overdrag;
    /**
     * Apply momentum from the pan gesture to the component when dragging finishes.
     * @default true
     */
    dragMomentum?: boolean;
    /**
     * Callback that fires when dragging starts
     */
    onDragStart?: (e: MouseEvent | TouchEvent) => void;
    /**
     * Callback that fires when dragging ends
     */
    onDragEnd?: (e: MouseEvent | TouchEvent) => void;
    /**
     * Callback that fires a drag direction is determined
     */
    onDirectionLock?: (axis: string) => void;
}

declare type Easing = CubicBezier | "linear" | "easeIn" | "easeOut" | "easeInOut" | "circIn" | "circOut" | "circInOut" | "backIn" | "backOut" | "backInOut" | "anticipate" | EasingFunction;

declare type EasingFunction = (v: number) => number;

declare type EventHandler = (event: Event, info: EventInfo) => void;

declare interface EventInfo {
    point: Point;
}

declare type GestureHandlers = PanHandlers & TapHandlers & HoverHandlers;

declare type HoverHandler = (event: Event) => void;

declare interface HoverHandlers {
    /**
     * Properties or variant label to animate to while the hover gesture is recognised
     */
    hover?: string | TargetAndTransition;
    /**
     * Callback that fires when pointer starts hovering over the component
     */
    onHoverStart?: HoverHandler;
    /**
     * Callback that fires when pointer stops hovering over the component
     */
    onHoverEnd?: HoverHandler;
}

declare type HTMLElements = UnionStringArray<typeof htmlElements>;

declare type HTMLMotionComponents = {
    [K in HTMLElements]: ComponentType<Omit<UnwrapFactory<ReactHTML[K]>, "style"> & MotionProps>;
};

declare type Inertia = {
    type: "inertia";
    modifyTarget?: (v: number) => number;
    bounceStiffness?: number;
    bounceDamping?: number;
    power?: number;
    timeConstant?: number;
    restDelta?: number;
    min?: number;
    max?: number;
    from?: number | string;
    velocity?: number;
    delay?: number;
};

declare type Just = {
    type: "just";
    from?: number | string;
    delay?: number;
    velocity?: number;
};

declare type Keyframes = {
    type: "keyframes";
    values: number[] | string[];
    easings?: Easing[];
    easeAll?: Easing;
    elapsed?: number;
    duration?: number;
    loop?: number;
    flip?: number;
    yoyo?: number;
    from?: number | string;
    to?: number | string;
    velocity?: number;
    delay?: number;
};

declare type MakeMotion<T> = {
    [K in keyof T]: T[K] | MotionValue<T[K]>;
} & MotionStyleProperties;

/**
 * HTML & SVG components, optimised for use with gestures and animation.
 *
 * @public
 */
export declare const motion: MotionComponents;

declare type MotionComponents = CustomMotionComponent & HTMLMotionComponents & SVGMotionComponents;

/**
 * @internal
 */
export declare const MotionPluginContext: React.Context<Partial<MotionPlugins>>;

declare interface MotionPluginProps extends MotionPlugins {
    children?: ReactNode;
}

/**
 * @internal
 * @internalremarks For now I think this should remain a private API for our own use
 * until we can figure out a nicer way of allowing people to add these
 */
export declare function MotionPlugins({ children, ...props }: MotionPluginProps): JSX.Element;

export declare interface MotionPlugins {
    transformPagePoint: (point: Point) => Point;
}

declare interface MotionProps extends GestureHandlers, DraggableProps {
    /**
     * Supports `MotionValue`s and separate `transform` values.
     */
    style?: MotionStyle;
    /**
     * Either properties to animate to, variant label, array of variant labels, or `AnimationController`
     */
    animate?: AnimationGroupControls | TargetAndTransition | VariantLabels;
    /**
     * Properties, variant label or array of variant labels to start in
     */
    initial?: Target | VariantLabels;
    /**
     * Object of variants
     */
    variants?: Variants;
    /**
     * Default transition
     */
    transition?: Transition;
    /**
     * Set to `false` to prevent inheriting variant changes from a parent `motion` component.
     * @default true
     */
    inherit?: boolean;
    /**
     * Set to `false` to block rendering the latest motion values on the component - can be used to temporarily disable animations for performance reasons.
     * @default true
     */
    render?: boolean;
    /**
     * Callback with latest motion values, fired max once per frame
     */
    onUpdate?: OnUpdate;
    /**
     * Callback when animation defined in `animate` is complete
     */
    onAnimationComplete?: () => void;
    [key: string]: any;
}

declare type MotionStyle = MakeMotion<CSSProperties>;

declare type MotionStyleProp = string | number | MotionValue;

declare interface MotionStyleProperties {
    x?: MotionStyleProp;
    y?: MotionStyleProp;
    z?: MotionStyleProp;
    rotate?: MotionStyleProp;
    rotateX?: MotionStyleProp;
    rotateY?: MotionStyleProp;
    rotateZ?: MotionStyleProp;
    scale?: MotionStyleProp;
    scaleX?: MotionStyleProp;
    scaleY?: MotionStyleProp;
    scaleZ?: MotionStyleProp;
    skew?: MotionStyleProp;
    skewX?: MotionStyleProp;
    skewY?: MotionStyleProp;
    originX?: MotionStyleProp;
    originY?: MotionStyleProp;
    originZ?: MotionStyleProp;
    perspective?: MotionStyleProp;
}

/**
 * @internal
 */
export declare class MotionValue<V = any> {
    private current;
    private prev;
    private timeDelta;
    private lastUpdated;
    private children?;
    private parent?;
    private updateSubscribers?;
    private renderSubscribers?;
    private transformer?;
    private controller?;
    private canTrackVelocity;
    constructor(init: V, { transformer, parent }?: Config<V>);
    addChild(config: Config<V>): MotionValue<V>;
    removeChild(child: MotionValue): void;
    subscribeTo(subscriptions: Set<Subscriber<V>>, subscription: Subscriber<V>): () => boolean;
    addUpdateSubscription(subscription: Subscriber<V>): () => boolean;
    addRenderSubscription(subscription: Subscriber<V>): () => boolean;
    set(v: V, render?: boolean): void;
    notifySubscriber: (subscriber: Subscriber<V>) => void;
    scheduleVelocityCheck: () => import("framesync/lib/types").Process;
    velocityCheck: ({ timestamp }: FrameData) => void;
    setChild: (child: MotionValue<any>) => void;
    get(): V;
    getVelocity(): number;
    control(controller: ActionFactory, config: PopmotionTransitionProps, transformer?: Transformer<V>): Promise<{}>;
    stop(): void;
    destroy(): void;
}

declare class MotionValuesMap {
    private hasMounted;
    private styler;
    private onUpdate?;
    private shouldRender;
    private values;
    private unsubscribers;
    has(key: string): boolean;
    set(key: string, value: MotionValue): void;
    get<Value>(key: string): MotionValue<Value> | undefined;
    get<Value>(key: string, defaultValue: Value): MotionValue<Value>;
    forEach(callback: (value: MotionValue, key: string) => void): void;
    bindValueToStyler(key: string, value: MotionValue): void;
    setOnUpdate(onUpdate?: OnUpdate): void;
    setShouldRender(render: boolean): void;
    mount(element: Element): void;
    unmount(): void;
}

declare type None = {
    type: false;
    from?: number | string;
    delay?: number;
    velocity?: number;
};

declare type Omit<T, K extends keyof T> = Pick<T, Exclude<keyof T, K>>;

declare type Omit_2<T, K extends keyof T> = Pick<T, Exclude<keyof T, K>>;

declare type OnUpdate = (v: Target) => void;

declare type Overdrag = boolean | number;

declare type PanHandler = (event: Event, info: PanInfo) => void;

declare interface PanHandlers {
    /**
     * Callback when the pan gesture is recognised on this element
     */
    onPan?: PanHandler;
    /**
     * Callback when the pan gesture begins on this element
     */
    onPanStart?: PanHandler;
    /**
     * Callback when the pan gesture ends on this element
     */
    onPanEnd?: PanHandler;
}

declare interface PanInfo {
    point: Point;
    delta: Point;
    offset: Point;
    velocity: Point;
}

declare type Physics = {
    type: "physics";
    acceleration?: number;
    friction?: number;
    restSpeed?: number | false;
    from?: number | string;
    velocity?: number;
    delay?: number;
};

declare interface Point {
    x: number;
    y: number;
}

declare namespace Point {
    /** @alpha */
    const subtract: (a: Point, b: Point) => Point;
}

declare type PopmotionTransitionProps = Tween | Spring | Decay | Keyframes | Physics | Inertia | Just;

declare interface ScrollMotionValues {
    scrollX: MotionValue<number>;
    scrollY: MotionValue<number>;
    scrollXProgress: MotionValue<number>;
    scrollYProgress: MotionValue<number>;
}

declare type Spring = {
    type: "spring";
    stiffness?: number;
    damping?: number;
    mass?: number;
    restSpeed?: number;
    restDelta?: number;
    from?: number | string;
    to?: number | string;
    velocity?: number;
    delay?: number;
};

declare type Subscriber<T> = (v: T) => void;

declare type SVGElements = UnionStringArray<typeof svgElements>;

declare type SVGMotionComponents = {
    [K in SVGElements]: ComponentType<Omit<SVGAttributes<SVGElement>, "style"> & MotionProps>;
};

declare type TapHandler = (event: Event, session: TapInfo) => void;

declare interface TapHandlers {
    /**
     * Callback when the tap gesture successfully ends on this element
     */
    onTap?: TapHandler;
    /**
     * Callback when the tap gesture starts on this element
     */
    onTapStart?: TapHandler;
    /**
     * Callback when the tap gesture ends outside this element
     */
    onTapCancel?: TapHandler;
    /**
     * Properties or variant label to animate to while the component is pressed
     */
    tap?: string | TargetAndTransition;
}

declare interface TapInfo {
    point: Point;
}

declare type Target = CSSPropertiesWithoutTransition & {
    x?: number | string;
    y?: number | string;
    z?: number | string;
    rotate?: number | string;
    rotateX?: number | string;
    rotateY?: number | string;
    rotateZ?: number | string;
    scale?: number | string;
    scaleX?: number | string;
    scaleY?: number | string;
    scaleZ?: number | string;
    skew?: number | string;
    skewX?: number | string;
    skewY?: number | string;
    originX?: number | string;
    originY?: number | string;
    originZ?: number | string;
    pathLength?: number;
    pathSpacing?: number;
};

declare type TargetAndTransition = Target & {
    transition?: Transition;
    transitionEnd?: Target;
};

declare type TargetResolver = (props: any, current: Target, velocity: Target) => TargetAndTransition;

declare type Transformer<T> = (v: T) => T;

declare type Transformer_2 = (v: any) => any;

declare type TransformerOptions = {
    clamp: true;
};

declare type Transition = (TransitionOrchestration & TransitionDefinition) | (TransitionOrchestration & TransitionMap);

declare type TransitionDefinition = Tween | Spring | Decay | Keyframes | Physics | Inertia | Just | None;

declare type TransitionMap = TransitionOrchestration & {
    [key: string]: TransitionDefinition;
};

declare type TransitionOrchestration = {
    delay?: number;
    beforeChildren?: boolean;
    afterChildren?: boolean;
    delayChildren?: number;
    staggerChildren?: number;
    staggerDirection?: 1 | -1;
};

declare type Tween = {
    type?: "tween";
    duration?: number;
    ease?: Easing;
    elapsed?: number;
    loop?: number;
    flip?: number;
    yoyo?: number;
    from?: number | string;
    to?: number | string;
    velocity?: number;
    delay?: number;
};

declare type UnionStringArray<T extends string[]> = T[number];

declare type UnwrapFactory<F> = F extends DetailedHTMLFactory<infer P, any> ? P : never;

/**
 * If the provided value is a MotionValue, this returns the actual value, otherwise just the value itself
 *
 * @internal
 */
export declare function unwrapMotionValue<V>(value: V | MotionValue<V>): V;

/**
 * Manually start, stop and sequence animations on one or more `motion` components.
 *
 * @params variants - List of variants. Optional
 * @returns Animation controller with `start` and `stop` methods
 *
 * @public
 */
export declare function useAnimation(variants?: Variants, defaultTransition?: Transition): AnimationGroupControls;

/**
 * Cycles through a series of states.
 *
 * @param items - An array of the possible states
 * @param initialIndex - Index of initial state. Optional, defaults to 0
 * @returns [currentState, cycleState]
 *
 * @public
 */
export declare function useCycle<T>(items: T[], initialIndex?: number): [T, (i?: any) => void];

/**
 * Add pan and tap gesture recognition to an element.
 *
 * @param props
 * @param ref
 * @public
 */
export declare function useGestures<P extends GestureHandlers>(props: P, ref: RefObject<Element>): void;

/**
 * Create a `MotionValue`
 *
 * @param {number | string} init - The initial state of the `MotionValue`
 * @returns `MotionValue`
 *
 * @public
 */
export declare function useMotionValue<T>(init: T): MotionValue<T>;

/**
 *
 * @param handlers
 * @param ref
 * @internal
 */
export declare function usePanGesture(handlers: PanHandlers, ref: RefObject<Element>): undefined;

export declare function usePanGesture(handlers: PanHandlers): {
    onPointerDown: EventHandler;
};

/**
 * @param handlers
 * @internal
 */
export declare function useTapGesture(handlers: TapHandlers & ControlsProp): {
    onPointerDown: EventHandler;
};

export declare function useTapGesture(handlers: TapHandlers & ControlsProp, ref: RefObject<Element>): undefined;

/**
 * Creates a new `MotionValue` that takes the output of another `MotionValue` and transforms it with a function
 *
 * @param value - `MotionValue`
 * @param transform - Function that accepts the output of `value` and returns a new value.
 * @returns `MotionValue`
 *
 * @public
 */
export declare function useTransformedValue(value: MotionValue, transform: Transformer_2): MotionValue;

/**
 * Creates a new `MotionValue` that takes the output of another `MotionValue` and transforms between the `from` range to the `to` range.
 *
 * @remarks
 * Options:
 * - `clamp`: Default `true`. Clamps output to the provided array
 *
 * @param value - `MotionValue`
 * @param from - A linear series of numbers (either all increasing or decreasing)
 * @param to - A series of numbers, colors or strings. Must be the same length as `from`.
 * @param options - Options
 * @returns `MotionValue`
 *
 * @public
 */
export declare function useTransformedValue(value: MotionValue<number>, from: number[], to: any[], options?: TransformerOptions): MotionValue;

/**
 * Provides `MotionValue`s that update when the viewport scrolls.
 *
 * @remarks
 *
 * This makes it possible to transform viewport scrolls into other values.
 *
 * Four `MotionValue`s are returned:
 *
 *  - `scrollX`/`scrollY`: The x/y scroll offset in pixels.
 *  - `scrollXProgress`/`scrollYProgress`: The x/y scroll offset as a progress value between `0` and `1`.
 *
 * ```jsx
 * const { scrollX } = useViewportScrollValues()
 * ```
 *
 * @returns `{ scrollX, scrollY, scrollXProgress, scrollYProgress }` `MotionValue`s
 *
 * @public
 */
export declare function useViewportScrollValues(): ScrollMotionValues;

declare type Variant = TargetAndTransition | TargetResolver;

declare type VariantLabels = string | string[];

declare type Variants = {
    [key: string]: Variant;
};

export { }
